module RotCig_mod

  use rotcig_types

  use fft_data_mod

  implicit none

  integer, external ::  sreed,srite,sseek_block

  logical linear_interp
  integer, private :: n_z,n_mx,n_z_pos,n_mx_pos
  integer, private :: n_z_pad,n_mx_pad,n_z_pad_pos,n_mx_pad_pos
  integer, private :: n_hx,n_hz,n_hdip,n_hrot,n_hpar
  integer, private :: n_bytes_panel
  integer, private :: n_bytes_nz_cmpx,n_bytes_hdip_cmpx,n_bytes_hpar_cmpx
  integer, private :: n_threads
  real, private :: d_z,d_mx,d_hx,d_hz,d_hdip,d_hrot,d_hpar
  real, private :: z_0,mx_0,hx_0,hz_0,hdip_0,hrot_0,hpar_0
  real, private :: kz_0,kmx_0,d_kz,d_kmx
  real, private :: pi,sin_rot,cos_rot
  real, private :: shift_z,shift_mx

  real, allocatable, dimension (:), private :: k_mx,k_z
  real, allocatable, dimension (:,:), private :: panel,panel_rot
  complex, allocatable, dimension (:,:), private :: temp_z,temp_x
  complex, allocatable, dimension (:,:), private :: k_panel_out,k_panel_inp
  complex, allocatable, dimension (:,:), private :: k_panel_dip,k_panel_par
  type (fft_type), private :: fft_nz_id,fft_nmx_id

contains

  function RotCig_init(rotcig_par)  result(status)
    type (rotcig_param_type) rotcig_par
    integer status

    integer i_k_mx,i_k_z,ierr,i_z
    real kz,kmx

    status=0

    linear_interp=rotcig_par%linear_interp

    shift_z=rotcig_par%rot_center_z- rotcig_par%z_0
    shift_mx=rotcig_par%rot_center_mx- rotcig_par%mx_0


    n_z=rotcig_par%n_z
    n_z_pad=rotcig_par%n_z_pad
    n_z_pos=rotcig_par%n_z/2
    n_z_pad_pos=rotcig_par%n_z_pad/2
    if(mod(n_z_pad,2) /=0) then
      write(0,*)'n_z_pad=',n_z_pad
      call seperr ("n_z_pad must be even")
    end if


    n_mx=rotcig_par%n_mx
    n_mx_pad=rotcig_par%n_mx_pad
    n_mx_pos=rotcig_par%n_mx/2
    n_mx_pad_pos=rotcig_par%n_mx_pad/2
    if(mod(n_mx_pad,2) /=0) then
      write(0,*)'n_mx_pad=',n_mx_pad
      call seperr ("n_mx_pad must be even")
    end if

    n_hx=rotcig_par%n_hx
    n_hz=rotcig_par%n_hz
    n_hrot=rotcig_par%n_hrot
    n_hdip=rotcig_par%n_hdip
    n_hpar=rotcig_par%n_hpar

    d_z=rotcig_par%d_z
    d_mx=rotcig_par%d_mx
    d_hx=rotcig_par%d_hx
    d_hz=rotcig_par%d_hz
    d_hrot=rotcig_par%d_hrot
    d_hdip=rotcig_par%d_hdip
    d_hpar=rotcig_par%d_hpar

    z_0=rotcig_par%z_0
    mx_0=rotcig_par%mx_0
    hx_0=rotcig_par%hx_0
    hz_0=rotcig_par%hz_0
    hrot_0=rotcig_par%hrot_0
    hdip_0=rotcig_par%hdip_0
    hpar_0=rotcig_par%hpar_0

    n_bytes_panel=n_z*n_mx*4
    n_bytes_nz_cmpx=(n_z_pad_pos+1)*8
    n_bytes_hdip_cmpx=(n_z_pad_pos+1)*n_hdip*8
    n_bytes_hpar_cmpx=(n_z_pad_pos+1)*n_hpar*8

    kz=0.
    kmx_0=0.
    pi=acos(-1.)
    d_kz=2.*pi/(d_z*FLOAT(n_z_pad))
    d_kmx=2.*pi/(d_mx*FLOAT(n_mx_pad))


    n_threads=rotcig_par%n_threads
    if(rotcig_par%create_cig_hpar) then
      if(mod(n_hdip,n_threads) /=0) then
        write(0,*)'n_threads=',n_threads
        write(0,*)'n_hdip=',n_hdip
        call seperr ("n_hdip be multiple of n_threads")
      end if
      if(mod(n_hpar,n_threads) /=0) then
        write(0,*)'n_threads=',n_threads
        write(0,*)'n_hpar=',n_hpar
        call seperr ("n_hpar be multiple of n_threads")
      end if
    else
      if(mod(n_hx,n_threads) /=0) then
        write(0,*)'n_threads=',n_threads
        write(0,*)'n_hx=',n_hrot
        call seperr ("n_hx be multiple of n_threads")
      end if
      if(mod(n_hz,n_threads) /=0) then
        write(0,*)'n_threads=',n_threads
        write(0,*)'n_hz=',n_hrot
        call seperr ("n_hz be multiple of n_threads")
      end if
      if(mod(n_hrot,n_threads) /=0) then
        write(0,*)'n_threads=',n_threads
        write(0,*)'n_hrot=',n_hrot
        call seperr ("n_hrot be multiple of n_threads")
      end if
     end if

    allocate (k_z(n_z_pad),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of k_z ")
    allocate (k_mx(n_mx_pad),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of k_mx ")

    allocate (panel(n_z,n_mx),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of panel ")
    allocate (panel_rot(n_z,n_mx),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of panel_rot ")

    allocate (temp_z(n_z_pad,n_threads),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of temp_z ")

    allocate (temp_x(n_mx_pad,n_threads),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of temp_x ")

    allocate (k_panel_out(n_z_pad_pos+1,n_mx_pad),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of k_panel_out ")

    if(rotcig_par%create_cig_hpar) then
      allocate (k_panel_dip(n_z_pad_pos+1,n_hdip),stat=ierr)  
      if(ierr .ne. 0) call seperr("Problems with allocation of k_panel_dip ")
      allocate (k_panel_par(n_z_pad_pos+1,n_hpar),stat=ierr)  
      if(ierr .ne. 0) call seperr("Problems with allocation of k_panel_par ")
    else
      allocate (k_panel_inp(n_z_pad,n_mx_pad),stat=ierr)  
      if(ierr .ne. 0) call seperr("Problems with allocation of k_panel_inp ")
    end if

    do i_k_z=1,n_z_pad
       kz  = i_k_z - 1.
       if(kz <= float(n_z_pad/2)) then
          k_z(i_k_z) =  kz_0+d_kz* kz
       else
          k_z(i_k_z) = -kz_0+d_kz*(kz-n_z_pad)
       end if
    end do

    do i_k_mx=1,n_mx_pad
       kmx  = i_k_mx - 1.
       if(kmx <= float(n_mx_pad/2)) then
          k_mx(i_k_mx) =  kmx_0+d_kmx* kmx
       else
          k_mx(i_k_mx) = -kmx_0+d_kmx*(kmx-n_mx_pad)
       end if
    end do


    fft_nz_id=init_fft_1d(rotcig_par%n_z_pad,fast_init=.false.)
    if(rotcig_par%debug1) write(0,*)'scale (nz_pad)=',fft_nz_id%scale
    status=fft_nz_id%ierr

    fft_nmx_id=init_fft_1d(rotcig_par%n_mx_pad,fast_init=.false.)
    if(rotcig_par%debug1) write(0,*)'scale (nmx_pad)=',fft_nmx_id%scale
    status=fft_nmx_id%ierr

    sin_rot=sin(pi*rotcig_par%out_dip_center/180.)
    cos_rot=cos(pi*rotcig_par%out_dip_center/180.)

     if(rotcig_par%debug1) then
write(0,*)'ierr (nmx)=',fft_nmx_id%ierr
write(0,*)'status=',status
write(0,*)'n_hdip=',n_hdip
write(0,*)'n_hpar=',n_hpar
write(0,*)'n_hrot=',n_hrot
write(0,*)'hdip_0=',hdip_0
write(0,*)'hpar_0=',hpar_0
write(0,*)'hrot_0=',hrot_0
write(0,*)'d_hdip=',d_hdip
write(0,*)'d_hpar=',d_hpar
write(0,*)'d_hrot=',d_hrot
write(0,*)'n_z=',n_z
write(0,*)'n_mx=',n_mx
write(0,*)'n_z_pad=',n_z_pad
write(0,*)'n_mx_pad=',n_mx_pad
write(0,*)'n_z_pos=',n_z_pos
write(0,*)'n_mx_pos=',n_mx_pos
write(0,*)'n_z_pad_pos=',n_z_pad_pos
write(0,*)'n_mx_pad_pos=',n_mx_pad_pos
write(0,*)'d_kz=',d_kz
write(0,*)'d_kmx=',d_kmx
write(0,*)'n_bytes_panel=',n_bytes_panel
write(0,*)'n_bytes_nz_cmpx=',n_bytes_nz_cmpx
write(0,*)'n_bytes_hdip_cmpx=',n_bytes_hdip_cmpx
write(0,*)'n_bytes_hpar_cmpx=',n_bytes_hpar_cmpx
write(0,*)'sin_rot=',sin_rot
write(0,*)'cos_rot=',cos_rot
write(0,*)'shift_z=',shift_z
write(0,*)'shift_mx=',shift_mx
     end if


  end function RotCig_init

  function RotCig_oper(rotcig_par) result(status)

    type (rotcig_param_type) rotcig_par
    integer status

    logical hz_rot_inside,hx_rot_inside,hdip_inside
    integer ret_sseek
    integer i_hrot,i_hz_rot,i_hx_rot,i_hpar,i_hdip,i_mx,i_z
    real h_rot,hz_rot,hx_rot,hz_min,hz_max,hx_min,hx_max
    real w_hz,w_hx,w_hz_rot,w_hx_rot,sin_rot_map,cos_rot_map
    real hdip_min,hdip_max,h_par,h_dip
    real tan_dip,dip,diff_dip,cos_diff,w_hdip

    if(rotcig_par%create_cig_hpar) then
    !{

      if(rotcig_par%debug2) write(0,*)'Beginning transforming cig_hdip'
      status= RotCig_transf( &
           .true.,rotcig_par%tag_cig_hdip,rotcig_par%tag_cig_hdip_transf,n_hdip,rotcig_par)
      if(rotcig_par%debug2) write(0,*)'Ending transforming cig_hdip'

      if(rotcig_par%debug2) write(0,*)'Beginning transforming offsets'

      hdip_min=hdip_0
      hdip_max=hdip_0+(n_hdip-1)*d_hdip

      ret_sseek=sseek_block( &
            trim(rotcig_par%tag_cig_hdip_transf),0,n_bytes_hdip_cmpx,0)
      if (ret_sseek .ne. 0) then
      !{
        write(0,*)'ret_sseek=',ret_sseek
        write(0,*)'tag_cig_hdip_transf=',trim(rotcig_par%tag_cig_hdip_transf)
        call seperr("Error block seeking cig_hdip_transf in input")
      !}
      end if
  
      do i_mx=1,n_mx_pad
      !{

        if(sreed(trim(rotcig_par%tag_cig_hdip_transf),k_panel_dip,n_bytes_hdip_cmpx)/= &
            n_bytes_hdip_cmpx) then
          write(0,*)'tag_cig_hdip_transf=',trim(rotcig_par%tag_cig_hdip_transf)
          call seperr('Problems with reading tag_cig_hdip_transf file')
        end if

        do i_hpar=1,n_hpar
        !{
          h_par=hpar_0+(i_hpar-1)*d_hpar
          if(rotcig_par%debug2) then
            if(mod(i_mx-2,n_mx_pad/4) ==0 ) then
              write(0,*)'i_mx=',i_mx
              write(0,*)'h_par=',h_par
            end if
          end if
          do i_z=1,n_z_pad_pos+1
          !{

            if(i_z /= 1) then
              tan_dip=k_mx(i_mx)/k_z(i_z)
              dip=-atan(tan_dip)
            else
              dip=pi
            end if
            diff_dip=dip-(pi*rotcig_par%inp_dip_center)/180.
            cos_diff=cos(diff_dip)

            if(abs(cos_diff) > 0.) then
              h_dip=h_par/cos_diff
              hdip_inside=((h_dip >= hdip_min) .and. (h_dip < hdip_max))
            else
              hdip_inside=.false.
            end if

            if(rotcig_par%debug2) then
              if(mod(i_mx-2,n_mx_pad/4) ==0 ) then
                if(mod(i_z-2,n_z_pad_pos/4) ==0 ) then
                  write(0,*)'i_z=',i_z
                  write(0,*)'dip=',dip*180/pi
                  write(0,*)'diff_dip=',diff_dip*180/pi
                  write(0,*)'cos_diff=',cos_diff
                  write(0,*)'h_dip=',h_dip
                  write(0,*)'hdip_inside=',hdip_inside
                end if
              end if
            end if

            if(hdip_inside) then

              i_hdip=floor((h_dip-hdip_0)/d_hdip) +1
              w_hdip=     ((h_dip-hdip_0)/d_hdip) +1  - i_hdip

              if(rotcig_par%debug2) then
                if(mod(i_mx-2,n_mx_pad/4) ==0 ) then
                  if(mod(i_z-2,n_z_pad_pos/4) ==0 ) then
                    write(0,*)'i_hdip=',i_hdip
                    write(0,*)'w_hdip=',w_hdip
                  end if
                end if
              end if

              k_panel_par(i_z,i_hpar)= (1.-w_hdip)*k_panel_dip(i_z,i_hdip  ) + &
                                       (   w_hdip)*k_panel_dip(i_z,i_hdip+1)
             else
              k_panel_par(i_z,i_hpar)= cmplx(0.,0.)
             end if

          !}
          end do
        !}
        end do


        if(srite(trim(rotcig_par%tag_cig_hpar_transf),k_panel_par,n_bytes_hpar_cmpx)/= &
            n_bytes_hpar_cmpx) then
          write(0,*)'tag_cig_hpar_transf=',trim(rotcig_par%tag_cig_hpar_transf)
          call seperr('Problems with writing tag_cig_hpar_transf file')
        end if

      !}
      end do

      if(rotcig_par%debug2) write(0,*)'Ending transforming offsets'

      if(rotcig_par%debug2) write(0,*)'Beginning transforming cig_hpar'
      status= RotCig_transf( &
           .false.,rotcig_par%tag_cig_hpar_transf,rotcig_par%tag_cig_hpar,n_hpar,rotcig_par)
      if(rotcig_par%debug2) write(0,*)'Ending transforming cig_hpar'

    !}
    else
    !{
      if(rotcig_par%cig_hz) then
      !{
        if(rotcig_par%debug2) write(0,*)'Beginning rotating cig_hz'
        status= RotCig_rot( &
             rotcig_par%tag_cig_hz,rotcig_par%tag_cig_hz_rot,n_hz,rotcig_par)
        if(rotcig_par%debug2) write(0,*)'Ending rotating cig_hz'
  
      !}
      end if
      if(rotcig_par%cig_hx) then
      !{
        if(rotcig_par%debug2) write(0,*)'Beginning rotating cig_hx'
        status= RotCig_rot( &
             rotcig_par%tag_cig_hx,rotcig_par%tag_cig_hx_rot,n_hx,rotcig_par)
        if(rotcig_par%debug2) write(0,*)'Ending rotating cig_hx'
  
      !}
      end if
  
      hz_min=hz_0
      hz_max=hz_0+(n_hz-1)*d_hz
      hx_min=hx_0
      hx_max=hx_0+(n_hx-1)*d_hx

      if(rotcig_par%no_h_mapping) then
       sin_rot_map=1.
       cos_rot_map=1.
      else
       sin_rot_map=sin_rot
       cos_rot_map=cos_rot
      end if
  
      do i_hrot=1,n_hrot
      !{
         panel_rot=0.
         h_rot=hrot_0+(i_hrot-1)*d_hrot
  
  
         if(rotcig_par%cig_hz) then
         !{
           if(abs(sin_rot_map) > 0.) then
             hz_rot=h_rot/sin_rot_map
             hz_rot_inside=((hz_rot > hz_min) .and. (hz_rot < hz_max))
           else
             hz_rot_inside=.false.
           end if
         !}
         else
         !{
           hz_rot_inside=.false.
         !}
         end if
  
         if(rotcig_par%cig_hx) then
         !{
           if(abs(cos_rot_map) > 0.) then
             hx_rot=h_rot/cos_rot_map
             hx_rot_inside=((hx_rot > hx_min) .and. (hx_rot < hx_max))
           else
             hx_rot_inside=.false.
           end if
         !}
         else
         !{
           hx_rot_inside=.false.
         !}
         end if
  
         if(hz_rot_inside .and. hx_rot_inside) then
           w_hz=sin_rot**2
           w_hx=cos_rot**2
         else
           if(hz_rot_inside) then
             w_hz=1.
             w_hx=0.
           else
             if(hx_rot_inside) then
               w_hz=0.
               w_hx=1.
             end if
           end if
         end if
  
  
        if(rotcig_par%debug2) then
          write(0,*)'h_rot=',h_rot
          write(0,*)'hz_rot=',hz_rot
          write(0,*)'hx_rot=',hx_rot
          write(0,*)'hz_rot_inside=',hz_rot_inside
          write(0,*)'hx_rot_inside=',hx_rot_inside
          write(0,*)'w_hz=',w_hz
          write(0,*)'w_hx=',w_hx
        end if
  
        if(hz_rot_inside) then
        !{
          i_hz_rot=floor((hz_rot-hz_0)/d_hz)
          w_hz_rot=(hz_rot-hz_0)/d_hz - i_hz_rot
  
         if(rotcig_par%debug2) then
           write(0,*)'i_hz_rot=',i_hz_rot
           write(0,*)'w_hz_rot=',w_hz_rot
         end if
  
          ret_sseek=sseek_block( &
                trim(rotcig_par%tag_cig_hz_rot),i_hz_rot,n_bytes_panel,0)
          if (ret_sseek .ne. i_hz_rot) then
          !{
            write(0,*)'ret_sseek=',ret_sseek
            write(0,*)'tag_cig_hz_rot=',trim(rotcig_par%tag_cig_hz_rot)
            call seperr("Error block seeking cig_hz_rot in input")
          !}
          end if
    
          if(sreed(trim(rotcig_par%tag_cig_hz_rot),panel,n_bytes_panel)/= &
              n_bytes_panel) then
            write(0,*)'tag_cig_hz_rot=',trim(rotcig_par%tag_cig_hz_rot)
            call seperr('Problems with reading cig_hz_rot file')
          end if
  
          panel_rot=panel_rot + w_hz*(1.-w_hz_rot)*panel
  
          if(sreed(trim(rotcig_par%tag_cig_hz_rot),panel,n_bytes_panel)/= &
              n_bytes_panel) then
            write(0,*)'tag_cig_hz_rot=',trim(rotcig_par%tag_cig_hz_rot)
            call seperr('Problems with reading cig_hz_rot file')
          end if
          panel_rot=panel_rot + w_hz*(w_hz_rot)*panel
        !}
        end if
  
        if(hx_rot_inside) then
        !{
          i_hx_rot=floor((hx_rot-hx_0)/d_hx)
          w_hx_rot=(hx_rot-hx_0)/d_hx - i_hx_rot
  
         if(rotcig_par%debug2) then
           write(0,*)'i_hx_rot=',i_hx_rot
           write(0,*)'w_hx_rot=',w_hx_rot
         end if
  
          ret_sseek=sseek_block( &
                trim(rotcig_par%tag_cig_hx_rot),i_hx_rot,n_bytes_panel,0)
          if (ret_sseek .ne. i_hx_rot) then
          !{
            write(0,*)'ret_sseek=',ret_sseek
            write(0,*)'tag_cig_hx_rot=',trim(rotcig_par%tag_cig_hx_rot)
            call seperr("Error block seeking cig_hx_rot in input")
          !}
          end if
    
          if(sreed(trim(rotcig_par%tag_cig_hx_rot),panel,n_bytes_panel)/= &
              n_bytes_panel) then
            write(0,*)'tag_cig_hx_rot=',trim(rotcig_par%tag_cig_hx_rot)
            call seperr('Problems with reading cig_hx_rot file')
          end if
  
          panel_rot=panel_rot + w_hx*(1.-w_hx_rot)*panel
  
          if(sreed(trim(rotcig_par%tag_cig_hx_rot),panel,n_bytes_panel)/= &
              n_bytes_panel) then
            write(0,*)'tag_cig_hx_rot=',trim(rotcig_par%tag_cig_hx_rot)
            call seperr('Problems with reading cig_hx_rot file')
          end if
          panel_rot=panel_rot + w_hx*(w_hx_rot)*panel
  
        !}
        end if
  
        if(srite(trim(rotcig_par%tag_cig_hrot),panel_rot,n_bytes_panel)/= &
            n_bytes_panel) then
          write(0,*)'tag_cig_hrot=',trim(rotcig_par%tag_cig_hrot)
          call seperr('Problems with writing cig_hrot file')
        end if
  
      !}
      end do 

    !}
    end if
  
  
      status=0
    return	
  end function RotCig_oper
!

  function RotCig_rot(tag_cig,tag_cig_rot,n_h,rotcig_par) result(status)

    character(len=128) :: tag_cig,tag_cig_rot
    integer n_h
    type (rotcig_param_type) rotcig_par
    integer status

    logical inside_kz,inside_kmx
    integer i_z,i_mx,i_h,i_kz_inp,i_kmx_inp,i_thread
    real kz,kmx,kz_inp,kmx_inp,inv_d_kz,inv_d_kmx
    real w_z,w_mx

    inv_d_kz=1./d_kz
    inv_d_kmx=1./d_kmx

    do i_thread=1,n_threads
    !{
    do i_h=1,n_h,n_threads
    !{
      if(sreed(trim(tag_cig),panel,n_bytes_panel)/= &
          n_bytes_panel) then
        write(0,*)'tag_cig=',trim(tag_cig)
        call seperr('Problems with reading tag_cig file')
      end if

! FFT along z
      do i_mx=1,n_mx
      !{
        temp_z(1:n_z,i_thread)=cmplx(panel(1:n_z,i_mx),0.)
        do i_z=n_z+1,n_z_pad
          temp_z(i_z,i_thread)=cmplx(0.,0.)
        end do
        call fft_1d_data_par(fft_nz_id,FFT_FORW,n_z_pad,i_thread,temp_z)

        do i_z=1,n_z_pad_pos+1
          k_panel_inp(i_z,i_mx)=temp_z(i_z,i_thread)
        end do

      !}
      end do 
      do i_mx=n_mx+1,n_mx_pad
      !{
        do i_z=1,n_z_pad_pos+1
          k_panel_inp(i_z,i_mx)=cmplx(0.,0.)
        end do
      !}
      end do 

      do i_z=n_z_pad_pos+2,n_z_pad
      !{
        k_panel_inp(i_z,:)=conjg(k_panel_inp(n_z_pad-i_z+2,:))
      !}
      end do

! FFT along mx
      do i_z=1,n_z_pad
      !{
        temp_x(:,i_thread)=k_panel_inp(i_z,:)
        call fft_1d_data_par(fft_nmx_id,FFT_FORW,n_mx_pad,i_thread,temp_x)
        k_panel_inp(i_z,:)=temp_x(:,i_thread)
!shift input
        do i_mx=1,n_mx_pad
        !{
          k_panel_inp(i_z,i_mx)=k_panel_inp(i_z,i_mx) * &
           cexp(cmplx(0,-k_z(i_z)*shift_z-k_mx(i_mx)*shift_mx))
        !}
        end do 
      !}
      end do 


! remap input 
      do i_mx=1,n_mx_pad
      !{
        do i_z=1,n_z_pad_pos+1
          temp_z(i_z+n_z_pad_pos-1,i_thread)=k_panel_inp(i_z,i_mx)
        end do
        do i_z=n_z_pad_pos+2,n_z_pad
          temp_z(i_z-n_z_pad_pos-1,i_thread)=k_panel_inp(i_z,i_mx)
        end do
        k_panel_inp(1:n_z_pad,i_mx)=temp_z(1:n_z_pad,i_thread)
      !}
      end do
      do i_z=1,n_z_pad
      !{
        do i_mx=1,n_mx_pad_pos+1
          temp_x(i_mx+n_mx_pad_pos-1,i_thread)=k_panel_inp(i_z,i_mx)
        end do
        do i_mx=n_mx_pad_pos+2,n_mx_pad
          temp_x(i_mx-n_mx_pad_pos-1,i_thread)=k_panel_inp(i_z,i_mx)
        end do
        k_panel_inp(i_z,1:n_mx_pad)=temp_x(1:n_mx_pad,i_thread)
      !}
      end do

! rotate
      do i_mx=1,n_mx_pad
      !{
        kmx=k_mx(i_mx)
        do i_z=1,n_z_pad_pos+1
        !{
          kz=k_z(i_z)

          kz_inp=  kz*cos_rot + kmx*sin_rot
          kmx_inp=-kz*sin_rot + kmx*cos_rot

          if(linear_interp) then
          !{
            i_kz_inp=floor(kz_inp*inv_d_kz+float(n_z_pad_pos))
            i_kmx_inp=floor(kmx_inp*inv_d_kmx+float(n_mx_pad_pos))

            inside_kz=((i_kz_inp >= 1) .and.(i_kz_inp < n_z_pad))
            inside_kmx=((i_kmx_inp >= 1) .and.(i_kmx_inp < n_mx_pad))

            w_z=(kz_inp*inv_d_kz)+n_z_pad_pos - i_kz_inp
            w_mx=(kmx_inp*inv_d_kmx)+n_mx_pad_pos - i_kmx_inp

            if(inside_kz .and. inside_kmx) then
            !{

              k_panel_out(i_z,i_mx)= &
                (1-w_z)*(1-w_mx)*k_panel_inp(i_kz_inp  ,i_kmx_inp  ) + &
                (1-w_z)*(  w_mx)*k_panel_inp(i_kz_inp  ,i_kmx_inp+1) + &
                (  w_z)*(1-w_mx)*k_panel_inp(i_kz_inp+1,i_kmx_inp  ) + &
                (  w_z)*(  w_mx)*k_panel_inp(i_kz_inp+1,i_kmx_inp+1)

            !}
            else
            !{
              if((i_kz_inp == 0) .and.(i_kmx_inp == 0)) then
              !{
                k_panel_out(i_z,i_mx)= &
                  (  w_z)*(  w_mx)*k_panel_inp(i_kz_inp+1,i_kmx_inp+1  )
              !}
              else
              !{
                if((i_kz_inp == n_z_pad) .and.(i_kmx_inp == 0)) then
                !{
                  k_panel_out(i_z,i_mx)= &
                    (1-w_z)*(  w_mx)*k_panel_inp(i_kz_inp  ,i_kmx_inp+1) 
                !}
                else
                !{
                  if((i_kz_inp == 0) .and.(i_kmx_inp == n_mx_pad)) then
                  !{
                    k_panel_out(i_z,i_mx)= &
                      (  w_z)*(1-w_mx)*k_panel_inp(i_kz_inp+1,i_kmx_inp  ) 
                  !}
                  else
                  !{
                    if((i_kz_inp == n_z_pad) .and.(i_kmx_inp == n_mx_pad))then
                      k_panel_out(i_z,i_mx)= &
                        (1-w_z)*(1-w_mx)*k_panel_inp(i_kz_inp  ,i_kmx_inp  )
                    else
                       k_panel_out(i_z,i_mx)=cmplx(0,0)
                    end if
                  !}
                  end if
                !}
                end if
              !}
              end if
            !}
            end if

          !}
          else
          !{
            i_kz_inp=nint(kz_inp*inv_d_kz+float(n_z_pad_pos))
            i_kmx_inp=nint(kmx_inp*inv_d_kmx+float(n_mx_pad_pos))

            inside_kz=((i_kz_inp >= 1) .and.(i_kz_inp <= n_z_pad))
            inside_kmx=((i_kmx_inp >= 1) .and.(i_kmx_inp <= n_mx_pad))

            if(inside_kz .and. inside_kmx) then
               k_panel_out(i_z,i_mx)=k_panel_inp(i_kz_inp,i_kmx_inp) 
            else
              k_panel_out(i_z,i_mx)=cmplx(0,0)
            end if
          !}
          end if

!shift output
          k_panel_out(i_z,i_mx)=k_panel_out(i_z,i_mx) * &
           cexp(cmplx(0,k_z(i_z)*shift_z+k_mx(i_mx)*shift_mx))

        !}
        end do
      !}
      end do 

! INV FFT along mx

      do i_z=1,n_z_pad_pos+1
      !{
        temp_x(1:n_mx_pad,i_thread)=k_panel_out(i_z,1:n_mx_pad)
        call fft_1d_data_par(fft_nmx_id,FFT_BACK,n_mx_pad,i_thread,temp_x)
        k_panel_out(i_z,1:n_mx_pad)=temp_x(1:n_mx_pad,i_thread)
      !}
      end do 

! NV FFT along z
      do i_mx=1,n_mx
      !{
        do i_z=1,n_z_pad_pos+1
        !{
          temp_z(i_z,i_thread)=k_panel_out(i_z,i_mx)
        !}
        end do
        do i_z=n_z_pad_pos+2,n_z_pad
        !{
          temp_z(i_z,i_thread)= conjg(k_panel_out(n_z_pad-i_z+2,i_mx))
        !}
        end do

        call fft_1d_data_par(fft_nz_id,FFT_BACK,n_z_pad,i_thread,temp_z)

        panel(1:n_z,i_mx)=real(temp_z(1:n_z,i_thread))

      !}
      end do 


      if(srite(trim(tag_cig_rot),panel,n_bytes_panel)/= &
          n_bytes_panel) then
        write(0,*)'tag_cig_rot=',trim(tag_cig_rot)
        call seperr('Problems with writing tag_cig_rot file')
      end if

    !}
    end do 
    !}
    end do 


    status=0
    return	

  end function RotCig_rot


  function RotCig_transf(fft_forward,tag_cig_inp,tag_cig_out,n_h,rotcig_par) &
              result(status)

    logical fft_forward
    character(len=128) :: tag_cig_inp,tag_cig_out
    integer n_h
    type (rotcig_param_type) rotcig_par
    integer status

    logical inside_kz,inside_kmx
    integer i_z,i_mx,i_h,i_kz_inp,i_kmx_inp,i_thread
    integer ret_sseek,i_sseek
    real kz,kmx,kz_inp,kmx_inp,inv_d_kz,inv_d_kmx
    real w_z,w_mx

    inv_d_kz=1./d_kz
    inv_d_kmx=1./d_kmx

    do i_thread=1,n_threads
    !{
    do i_h=1,n_h,n_threads
    !{


      if(fft_forward) then
      !{
        if(sreed(trim(tag_cig_inp),panel,n_bytes_panel)/= &
            n_bytes_panel) then
          write(0,*)'tag_cig=',trim(tag_cig_inp)
          call seperr('Problems with reading tag_cig_inp file')
        end if
  
  ! FFT along z
        do i_mx=1,n_mx
        !{
          temp_z(1:n_z,i_thread)=cmplx(panel(1:n_z,i_mx),0.)
          do i_z=n_z+1,n_z_pad
            temp_z(i_z,i_thread)=cmplx(0.,0.)
          end do
          call fft_1d_data_par(fft_nz_id,FFT_FORW,n_z_pad,i_thread,temp_z)
  
          do i_z=1,n_z_pad_pos+1
            k_panel_out(i_z,i_mx)=temp_z(i_z,i_thread)
          end do
  
        !}
        end do 
        do i_mx=n_mx+1,n_mx_pad
        !{
          do i_z=1,n_z_pad_pos+1
            k_panel_out(i_z,i_mx)=cmplx(0.,0.)
          end do
        !}
        end do 
  
  ! FFT along mx
        do i_z=1,n_z_pad_pos+1
        !{
          temp_x(:,i_thread)=k_panel_out(i_z,:)
          call fft_1d_data_par(fft_nmx_id,FFT_FORW,n_mx_pad,i_thread,temp_x)
          k_panel_out(i_z,:)=temp_x(:,i_thread)
        !}
        end do 

        do i_mx=1,n_mx_pad
        !{

          i_sseek=((i_mx-1)*n_h + (i_h-1))
          ret_sseek=sseek_block( &
                trim(tag_cig_out),i_sseek,n_bytes_nz_cmpx,0)
          if (ret_sseek .ne. i_sseek) then
          !{
            write(0,*)'ret_sseek=',ret_sseek
            write(0,*)'i_sseek=',i_sseek
            write(0,*)'tag_cig_out=',trim(tag_cig_out)
            call seperr("Error block seeking cig_out in output")
          !}
          end if
    
          do i_z=1,n_z_pad_pos+1
            temp_z(i_z,i_thread)=k_panel_out(i_z,i_mx)
          end do

          if(srite(trim(tag_cig_out),temp_z,n_bytes_nz_cmpx)/= &
              n_bytes_nz_cmpx) then
            write(0,*)'tag_cig_out=',trim(tag_cig_out)
            call seperr('Problems with writing tag_cig_out file')
          end if

        !}
        end do 

      !}
      else
      !{

        do i_mx=1,n_mx_pad
        !{

          i_sseek=((i_mx-1)*n_h + (i_h-1))
          ret_sseek=sseek_block( &
                trim(tag_cig_inp),i_sseek,n_bytes_nz_cmpx,0)
          if (ret_sseek .ne. i_sseek) then
          !{
            write(0,*)'ret_sseek=',ret_sseek
            write(0,*)'i_sseek=',i_sseek
            write(0,*)'tag_cig_out=',trim(tag_cig_inp)
            call seperr("Error block seeking cig_inp in output")
          !}
          end if

          if(sreed(trim(tag_cig_inp),temp_z,n_bytes_nz_cmpx)/= &
              n_bytes_nz_cmpx) then
            write(0,*)'tag_cig_inp=',trim(tag_cig_inp)
            call seperr('Problems with reading tag_cig_inp file')
          end if

          do i_z=1,n_z_pad_pos+1
            k_panel_out(i_z,i_mx)=temp_z(i_z,i_thread)
          end do

        !}
        end do 

! INV FFT along mx

        do i_z=1,n_z_pad_pos+1
        !{
          temp_x(1:n_mx_pad,i_thread)=k_panel_out(i_z,1:n_mx_pad)
          call fft_1d_data_par(fft_nmx_id,FFT_BACK,n_mx_pad,i_thread,temp_x)
          k_panel_out(i_z,1:n_mx_pad)=temp_x(1:n_mx_pad,i_thread)
        !}
        end do 

! NV FFT along z
        do i_mx=1,n_mx
        !{
          do i_z=1,n_z_pad_pos+1
          !{
            temp_z(i_z,i_thread)=k_panel_out(i_z,i_mx)
          !}
          end do
          do i_z=n_z_pad_pos+2,n_z_pad
          !{
            temp_z(i_z,i_thread)= conjg(k_panel_out(n_z_pad-i_z+2,i_mx))
          !}
          end do
  
          call fft_1d_data_par(fft_nz_id,FFT_BACK,n_z_pad,i_thread,temp_z)
  
          panel(1:n_z,i_mx)=real(temp_z(1:n_z,i_thread))
  
        !}
        end do 


        if(srite(trim(tag_cig_out),panel,n_bytes_panel)/= &
            n_bytes_panel) then
          write(0,*)'tag_cig_out=',trim(tag_cig_out)
          call seperr('Problems with writing tag_cig_out file')
        end if

      !}
      end if

    !}
    end do 
    !}
    end do 


    status=0
    return	

  end function RotCig_transf


subroutine fft_1d_data_par(fft_id,sign,n1,i_thread,data_par)
type (fft_type) fft_id
integer :: n1,i_thread
integer :: sign
complex, dimension (:,:) :: data_par

call fft_1d_data(fft_id,sign,n1,data_par(:,i_thread))
!
end subroutine fft_1d_data_par


end module RotCig_mod
